"""Tests for nepc.util.config.py"""
import os
from nepc.util import config


def test_user_home():
    """Verify whether user_home() returns a string
    that represents the correct path to the user's home directory"""
    assert isinstance(config.user_home(), str)
    assert config.user_home() == os.environ.get('HOME')


def test_nepc_home():
    """Verify whether nepc_home() returns a string
    that represents the correct path to the NEPC folder"""
    assert isinstance(config.nepc_home(), str)
    assert config.nepc_home() == os.environ.get('NEPC_HOME')



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'nepc'
copyright = '2020, NRL'
author = 'Paul Adamson'

# The full version, including alpha/beta/rc tags
#release = '0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.napoleon',
              'sphinx.ext.intersphinx']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'classic'
html_logo = 'fig/nepc_logo.png'

# -- Options for pdf output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
latex_logo = 'fig/nepc_logo.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Document __init__ methods -----------------------------------------------
#def skip(app, what, name, obj, would_skip, options):
    #if name == "__init__":
        #return False
    #return would_skip
#
#def setup(app):
    #app.connect("autodoc-skip-member", skip)

# -- intersphinx mappings ----------------------------------------------------

intersphinx_mapping = {'python': ('https://docs.python.org/3', None),
                       'matplotlib': ('https://matplotlib.org', None),
                       'numpy': ('https://numpy.org/doc/stable/', None)}

# -- Napoleon settings--------------------------------------------------------

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True



"""Builds the NEPC database"""
import argparse
import platform
import os
import mysql.connector
import pandas as pd
from nepc import nepc
from nepc.util import config
import time

NA_VALUES = []
PARSER = argparse.ArgumentParser(description='Build the NEPC database.')
PARSER.add_argument('--debug', action='store_true',
                    help='print additional debug info')
PARSER.add_argument('--test', action='store_true',
                    help='build test database')
ARGS = PARSER.parse_args()

if ARGS.debug:
    MAX_CS = 50
    MAX_RATE = 50
else:
    MAX_CS = 2000000
    MAX_RATE = 2000000

if ARGS.test:
    database = 'nepc_test'
    NEPC_DATA = config.nepc_home() + "/tests/data/"
    DIR_NAMES = ["/cs/n2/fict/",
                 "/cs/n2/fict_total/"]
else:
    database = 'nepc'
    NEPC_DATA = config.nepc_cs_home() + "/data/"
    DIR_NAMES = ["/cs/n2/itikawa/",
                 "/cs/n2/zipf/",
                 "/cs/n/zatsarinny/",
                 "/cs/n2/phelps/",
                 "/cs/n2/phelps_total/",
                 "/cs/n2/little/"]

T0 = time.time()

HOME = config.user_home()

def np_str(df, row, name):
    if df.iloc[row][name] == "\\N":
        return "Null"
    else:
        return df.iloc[row][name]


def print_table(table):
    """print a table
    ARGUMENT
    ========
    table: str
    """
    MYCURSOR.execute("SELECT * FROM " + table + ";")
    table_data = MYCURSOR.fetchall()
    for row in table_data:
        print(row)


def print_timestep(stage):
    """print stage and elapsed time

    ARGUMENT
    ========
    stage: str
    """
    current_time = time.time()
    elapsed = current_time-T0
    print("\n" + stage + ": " + str(round(elapsed, 2)) + " sec\n"
          "===============================================")


def insert_command(table_name, variable_list):
    variable_list_str = ", ".join(variable_list)
    data_format_str = ", ".join(["%s"]*len(variable_list))
    return ("INSERT INTO " + table_name +
            "(" + variable_list_str + ") " +
            "VALUES (" + data_format_str + ");")


####################
# Connect to MySQL #
####################

MYDB = mysql.connector.connect(
    host='localhost',
    option_files=HOME + '/.mysql/defaults'
)

MYCURSOR = MYDB.cursor()

if ARGS.debug:
    print_timestep("connected to MySQL")

##############################
# Create empty NEPC database #
##############################

MYCURSOR.execute("DROP DATABASE IF EXISTS `" + database + "`;")

MYCURSOR.execute("CREATE DATABASE IF NOT EXISTS `" + database + "` "
                 "CHARACTER SET utf8 "
                 "COLLATE utf8_general_ci;")

MYCURSOR.execute("SET default_storage_engine = INNODB;")

MYCURSOR.execute("use " + database + ";")

MYCURSOR.execute("CREATE TABLE `" + database + "`.`species`("
                 "`id` INT UNSIGNED NOT NULL ,"
                 "`name` VARCHAR(40) NOT NULL ,"
                 "`long_name` VARCHAR(100) NOT NULL ,"
                 "PRIMARY KEY(`id`)"
                 ");")

MYCURSOR.execute("CREATE TABLE `" + database + "`.`processes`( "
                 "`id` INT UNSIGNED NOT NULL , "
                 "`name` VARCHAR(40) NOT NULL, "
                 "`long_name` VARCHAR(240) NOT NULL, "
                 "`lhs` INT, "
                 "`rhs` INT, "
                 "`lhs_e` INT, "
                 "`rhs_e` INT, "
                 "`lhs_hv` INT, "
                 "`rhs_hv` INT, "
                 "`lhs_v` INT, "
                 "`rhs_v` INT, "
                 "`lhs_j` INT, "
                 "`rhs_j` INT, "
                 "PRIMARY KEY(`id`) "
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`states`( "
                 "`id` INT UNSIGNED NOT NULL, "
                 "`specie_id` INT UNSIGNED NOT NULL, "
                 "`name` VARCHAR(100) NOT NULL, "
                 "`long_name` VARCHAR(100) NOT NULL, "
                 "PRIMARY KEY(`id`), "
                 "INDEX `SPECIE_ID`(`specie_id` ASC), "
                 "CONSTRAINT `specie_id_STATES` FOREIGN KEY(`specie_id`) "
                 "REFERENCES `" + database + "`.`species`(`id`) "
                 "ON DELETE RESTRICT ON UPDATE CASCADE "
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`models`("
                 "	`model_id` INT UNSIGNED NOT NULL ,"
                 "	`name` VARCHAR(40) NOT NULL ,"
                 "	`long_name` VARCHAR(240) NOT NULL ,"
                 "	`ref` VARCHAR(40) NOT NULL ,"
                 "	PRIMARY KEY(`model_id`)"
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`cs`("
                 "	`cs_id` INT UNSIGNED NOT NULL AUTO_INCREMENT, "
                 "	`specie_id` INT UNSIGNED NOT NULL, "
                 "	`process_id` INT UNSIGNED NOT NULL, "
                 "	`units_e` DOUBLE NOT NULL,"
                 "	`units_sigma` DOUBLE NOT NULL,"
                 "	`ref` VARCHAR(1000),"
                 "	`lhsA_id` INT UNSIGNED NULL ,"
                 "	`lhsB_id` INT UNSIGNED NULL ,"
                 "	`rhsA_id` INT UNSIGNED NULL ,"
                 "	`rhsB_id` INT UNSIGNED NULL ,"
                 "	`threshold` DOUBLE NULL ,"
                 "	`wavelength` DOUBLE NULL ,"
                 "	`lhs_v` INT NULL ,"
                 "	`rhs_v` INT NULL ,"
                 "	`lhs_j` INT NULL ,"
                 "	`rhs_j` INT NULL ,"
                 "	`background` VARCHAR(10000) ,"
                 "	`lpu` DOUBLE NULL ,"
                 "	`upu` DOUBLE NULL ,"
                 "	PRIMARY KEY(`CS_ID`) ,"
                 "	INDEX `SPECIE_ID`(`specie_id` ASC) ,"
                 "	CONSTRAINT `SPECIE_ID_CS` FOREIGN KEY(`specie_id`)"
                 "	REFERENCES `" + database + "`.`species`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE,"
                 "	INDEX `PROCESS_ID`(`process_id` ASC) ,"
                 "	CONSTRAINT `PROCESS_ID_CS` FOREIGN KEY(`process_id`)"
                 "	REFERENCES `" + database + "`.`processes`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `LHSA_ID_CS` FOREIGN KEY(`lhsA_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `LHSB_ID_CS` FOREIGN KEY(`lhsB_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `RHSA_ID_CS` FOREIGN KEY(`rhsA_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `RHSB_ID_CS` FOREIGN KEY(`rhsB_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE "
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`csdata`("
                 "	`id` INT UNSIGNED NOT NULL AUTO_INCREMENT,"
                 "	`cs_id` INT UNSIGNED NOT NULL ,"
                 "	`e` DOUBLE NOT NULL ,"
                 "	`sigma` DOUBLE NOT NULL ,"
                 "	PRIMARY KEY(`id`) ,"
                 "	INDEX `cs_id`(`cs_id` ASC) ,"
                 "	CONSTRAINT `CS_ID_CSDATA` FOREIGN KEY(`cs_id`)"
                 "		REFERENCES `" + database + "`.`cs`(`cs_id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE "
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`rate`("
                 "	`rate_id` INT UNSIGNED NOT NULL AUTO_INCREMENT, "
                 "	`specie_id` INT UNSIGNED NOT NULL, "
                 "	`process_id` INT UNSIGNED NOT NULL, "
                 "	`ref` VARCHAR(1000),"
                 "	`lhsA_id` INT UNSIGNED NULL ,"
                 "	`lhsB_id` INT UNSIGNED NULL ,"
                 "	`rhsA_id` INT UNSIGNED NULL ,"
                 "	`rhsB_id` INT UNSIGNED NULL ,"
                 "	`threshold` DOUBLE NULL ,"
                 "	`wavelength` DOUBLE NULL ,"
                 "	`lhs_v` INT NULL ,"
                 "	`rhs_v` INT NULL ,"
                 "	`lhs_j` INT NULL ,"
                 "	`rhs_j` INT NULL ,"
                 "	`background` VARCHAR(10000), "
                 "      `form` VARCHAR(100) NOT NULL, "
                 "	PRIMARY KEY(`RATE_ID`) ,"
                 "	INDEX `SPECIE_ID`(`specie_id` ASC) ,"
                 "	CONSTRAINT `SPECIE_ID_RATE` FOREIGN KEY(`specie_id`)"
                 "	REFERENCES `" + database + "`.`species`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE,"
                 "	INDEX `PROCESS_ID`(`process_id` ASC) ,"
                 "	CONSTRAINT `PROCESS_ID_RATE` FOREIGN KEY(`process_id`)"
                 "	REFERENCES `" + database + "`.`processes`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `LHSA_ID_RATE` FOREIGN KEY(`lhsA_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `LHSB_ID_RATE` FOREIGN KEY(`lhsB_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `RHSA_ID_RATE` FOREIGN KEY(`rhsA_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE, "
                 "	CONSTRAINT `RHSB_ID_RATE` FOREIGN KEY(`rhsB_id`)"
                 "		REFERENCES `" + database + "`.`states`(`id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE "
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`ratedata`("
                 "	`id` INT UNSIGNED NOT NULL AUTO_INCREMENT,"
                 "	`rate_id` INT UNSIGNED NOT NULL ,"
                 "	`num` DOUBLE NOT NULL ,"
                 "	`constant` DOUBLE NOT NULL ,"
                 "	`lau` DOUBLE NOT NULL ,"
                 "	`uau` DOUBLE NOT NULL ,"
                 "	PRIMARY KEY(`id`) ,"
                 "	INDEX `rate_id`(`rate_id` ASC) ,"
                 "	CONSTRAINT `RATE_ID_RATEDATA` FOREIGN KEY(`rate_id`)"
                 "		REFERENCES `" + database + "`.`rate`(`rate_id`)"
                 "  ON DELETE RESTRICT ON UPDATE CASCADE "
                 ");"
                 )

MYDB.commit()

MYCURSOR.execute("CREATE TABLE `" + database + "`.`models2cs`("
                 "	`cs_id` INT UNSIGNED NOT NULL ,"
                 "	`model_id` INT UNSIGNED NOT NULL ,"
                 "	PRIMARY KEY pk_models2cs (cs_id, model_id)"
                 ");"
                 )

MYCURSOR.execute("CREATE TABLE `" + database + "`.`models2rate`("
                 "	`rate_id` INT UNSIGNED NOT NULL ,"
                 "	`model_id` INT UNSIGNED NOT NULL ,"
                 "	PRIMARY KEY pk_models2rate (rate_id, model_id)"
                 ");"
                 )

MYDB.commit()

if ARGS.debug:
    print_timestep("created empty NEPC database and all tables")

#############
# Load data #
#############

if ARGS.debug:
    print_timestep("starting to load data into processes table")

PROCESSES_VARIABLE_LIST = ["id", "name", "long_name", "lhs", "rhs",
                           "lhs_e", "rhs_e", "lhs_hv", "rhs_hv",
                           "lhs_v", "rhs_v", "lhs_j", "rhs_j"]
INSERT_COMMAND_PROCESSES = insert_command("processes",
                                          PROCESSES_VARIABLE_LIST)
PROCESSES_DTYPE = {"id": int,
                   "name": str,
                   "long_name": str,
                   "lhs": int,
                   "rhs": int,
                   "lhs_e": int,
                   "rhs_e": int,
                   "lhs_hv": int,
                   "rhs_hv": int,
                   "lhs_v": int,
                   "rhs_v": int,
                   "lhs_j": int,
                   "rhs_j": int}
PROCESSES_FILE = NEPC_DATA + "processes.tsv"
PROCESSES_DATA_LIST = list(pd.read_csv(PROCESSES_FILE,
                                       sep='\t',
                                       dtype=PROCESSES_DTYPE,
                                       na_values=NA_VALUES).itertuples(
                                           index=False,
                                           name=None))
MYCURSOR.executemany(INSERT_COMMAND_PROCESSES, PROCESSES_DATA_LIST)
if ARGS.debug:
    print_timestep("loaded data into processes table")
    # print_table("processes")

MODELS_VARIABLE_LIST = ["model_id", "name", "long_name", "ref"]
INSERT_COMMAND_MODELS = insert_command("models",
                                       MODELS_VARIABLE_LIST)
MODELS_DTYPE = {"model_id": int,
                "name": str,
                "long_name": str,
                "ref": str}
MODELS_FILE = NEPC_DATA + "models.tsv"
MODELS_DATA_LIST = list(pd.read_csv(MODELS_FILE,
                                    sep='\t',
                                    dtype=MODELS_DTYPE,
                                    na_values=NA_VALUES).itertuples(
                                        index=False,
                                        name=None))
if ARGS.debug:
    print_timestep("loading data into models table")
    print(MODELS_DATA_LIST)
MYCURSOR.executemany(INSERT_COMMAND_MODELS, MODELS_DATA_LIST)
if ARGS.debug:
    print_timestep("loaded data into models table")
    print_table("models")

SPECIES_VARIABLE_LIST = ["id", "name", "long_name"]
INSERT_COMMAND_SPECIES = insert_command("species",
                                        SPECIES_VARIABLE_LIST)
SPECIES_DTYPE = {"id": int,
                 "name": str,
                 "long_name": str}
SPECIES_FILE = NEPC_DATA + "species.tsv"
SPECIES_DATA_LIST = list(pd.read_csv(SPECIES_FILE,
                                     sep='\t',
                                     dtype=SPECIES_DTYPE,
                                     na_values=NA_VALUES).itertuples(
                                         index=False,
                                         name=None))
MYCURSOR.executemany(INSERT_COMMAND_SPECIES, SPECIES_DATA_LIST)
if ARGS.debug:
    print_timestep("loaded data into species table")
    # print_table("species")

STATES_VARIABLE_LIST = ["id", "name", "long_name", "specie"]
INSERT_COMMAND_STATES = ("INSERT INTO states "
                         "SET "
                         "  id = %s, "
                         "  name = %s, "
                         "  long_name = %s, "
                         "  specie_id = (SELECT id FROM " + database + ".species "
                         "              WHERE NAME = %s);")

STATES_DTYPE = {"id": int,
                "name": str,
                "long_name": str,
                "specie": str}

STATES_FILE = NEPC_DATA + "states.tsv"
STATES_DATA_LIST = list(pd.read_csv(STATES_FILE,
                                    sep='\t',
                                    dtype=STATES_DTYPE,
                                    na_values=NA_VALUES).itertuples(
                                        index=False,
                                        name=None))
MYCURSOR.executemany(INSERT_COMMAND_STATES, STATES_DATA_LIST)
if ARGS.debug:
    print_timestep("loaded data into states table")
    # print_table("states")

if ARGS.debug:
    print_timestep("starting to load cross section data")


if platform.node() == 'ppdadamsonlinux':
    CS_DAT_FILENAME = NEPC_DATA + "cs_datfile_prod.tsv"
else:
    CS_DAT_FILENAME = NEPC_DATA + "cs_datfile_local.tsv"

F_CS_DAT_FILE = open(CS_DAT_FILENAME, 'w')
F_CS_DAT_FILE.write("\t".join(["cs_id", "filename"]) + "\n")

CS_VARIABLE_LIST = ["cs_id", "specie_id", "process_id",
                    "units_e", "units_sigma",
                    "ref",
                    "lhsA_id", "lhsB_id", "rhsA_id", "rhsB_id",
                    "threshold",
                    "wavelength",
                    "lhs_v", "rhs_v", "lhs_j", "rhs_j", "background",
                    "lpu", "upu"]
CS_DTYPE = {"cs_id": int,
            "specie": str,
            "process": str,
            "units_e": float,
            "units_sigma": float,
            "ref": str,
            "lhs_a": str,
            "lhs_b": str,
            "rhs_a": str,
            "rhs_b": str,
            "threshold": float,
            "wavelength": float,
            "lhs_v": int,
            "rhs_v": int,
            "lhs_j": int,
            "rhs_j": int,
            "background": str,
            "lpu": float,
            "upu": float
            }
CSDATA_VARIABLE_LIST = ["id", "cs_id", "e", "sigma"]
CSDATA_DTYPE = {"csdata_id": int,
                "e_energy": float,
                "sigma": float}
MOD_DTYPE = {"model_name": str}

UPDATE_COMMAND_CS = ("INSERT INTO cs " +
                     "SET "
                     "  cs_id = %s, "
                     "  units_e = %s, "
                     "  units_sigma = %s, "
                     "  ref = %s, "
                     "  threshold = %s, "
                     "  wavelength = %s, "
                     "  lhs_v = %s, "
                     "  rhs_v = %s, "
                     "  lhs_j = %s, "
                     "  rhs_j = %s, "
                     "  background = %s, "
                     "  lpu = %s, "
                     "  upu = %s, "
                     "  specie_id = (SELECT id FROM " + database + ".species "
                     "    WHERE NAME=%s), "
                     "  process_id = (select id from " + database + ".processes "
                     "    WHERE NAME=%s), "
                     "  lhsA_id = (select id from " + database + ".states "
                     "    WHERE NAME=%s), "
                     "  lhsB_id = (select id from " + database + ".states "
                     "    WHERE NAME=%s), "
                     "  rhsA_id = (select id from " + database + ".states "
                     "    WHERE NAME=%s), "
                     "  rhsB_id = (select id from " + database + ".states "
                     "    WHERE NAME=%s);")

INSERT_COMMAND_CSDATA = insert_command("csdata", CSDATA_VARIABLE_LIST)

INSERT_COMMAND_MODELS2CS = ("INSERT INTO models2cs "
                            "SET "
                            " cs_id=%s, "
                            " model_id=(SELECT model_id "
                            "           FROM " + database + ".models "
                            "           WHERE NAME=%s);")

file_number = 1
for directoryname in DIR_NAMES:
    if ARGS.debug:
        print_timestep('entering directory: {}'.format(directoryname))
    directory = os.fsencode(NEPC_DATA + directoryname)
    for file in os.listdir(directory):
        if file_number >= MAX_CS:
            print("WARNING: only processed " + str(file_number) +
                  " cross section files. There appears to be addtional "
                  "data not processed.")
            break
        filename = os.fsdecode(file)
        if filename.endswith(".met") or filename.endswith(".mod"):
            continue
        else:
            # print("file_number: " + str(file_number))
            file_number = file_number + 1
            filename_wo_ext = filename.rsplit(".", 1)[0]
            mod_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".mod"])
            met_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".met"])

            dat_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".dat"])

            met_data = pd.read_csv(met_file,
                                   sep='\t',
                                   dtype=CS_DTYPE,
                                   na_values=NA_VALUES)
            dat_data = pd.read_csv(dat_file,
                                   sep='\t',
                                   dtype=CSDATA_DTYPE,
                                   na_values=NA_VALUES)

            F_CS_DAT_FILE.write(
                "\t".join([str(met_data.iloc[0]['cs_id']),
                           directoryname + str(filename_wo_ext)]) + "\n"
            )

            MYCURSOR.execute(UPDATE_COMMAND_CS,
                             (int(met_data.iloc[0]['cs_id']),
                              met_data.iloc[0]['units_e'],
                              met_data.iloc[0]['units_sigma'],
                              met_data.iloc[0]['ref'],
                              met_data.iloc[0]['threshold'],
                              met_data.iloc[0]['wavelength'],
                              int(met_data.iloc[0]['lhs_v']),
                              int(met_data.iloc[0]['rhs_v']),
                              int(met_data.iloc[0]['lhs_j']),
                              int(met_data.iloc[0]['rhs_j']),
                              met_data.iloc[0]['background'],
                              met_data.iloc[0]['lpu'],
                              met_data.iloc[0]['upu'],
                              np_str(met_data, 0, 'specie'),
                              np_str(met_data, 0, 'process'),
                              np_str(met_data, 0, 'lhs_a'),
                              np_str(met_data, 0, 'lhs_b'),
                              np_str(met_data, 0, 'rhs_a'),
                              np_str(met_data, 0, 'rhs_b')))
            dat_data.insert(1, 'cs_id', met_data.iloc[0]['cs_id'])
            dat_data_list = list(dat_data.itertuples(index=False,
                                                     name=None))
            MYCURSOR.executemany(INSERT_COMMAND_CSDATA, dat_data_list)

            if os.path.exists(mod_file):
                mod_data = pd.read_csv(mod_file,
                                       sep='\t',
                                       dtype=MOD_DTYPE,
                                       na_values=NA_VALUES)
                mod_data.insert(0, 'cs_id', met_data.iloc[0]['cs_id'])
                mod_data_list = list(mod_data.itertuples(index=False,
                                                         name=None))
                MYCURSOR.executemany(INSERT_COMMAND_MODELS2CS,
                                     mod_data_list)

            MYDB.commit()

MYDB.commit()
F_CS_DAT_FILE.close()

if ARGS.debug:
    print_timestep("loaded data into cs, csdata, and models2cs tables")
    # print_table("cs")

# DIR_NAMES = ["/data/rate/n2/peterson/"]
DIR_NAMES = []

if platform.node() == 'ppdadamsonlinux':
    RATE_DAT_FILENAME = NEPC_DATA + "rate_datfile_prod.tsv"
else:
    RATE_DAT_FILENAME = NEPC_DATA + "rate_datfile_local.tsv"

F_RATE_DAT_FILE = open(RATE_DAT_FILENAME, 'w')
F_RATE_DAT_FILE.write("\t".join(["rate_id", "filename"]) + "\n")

RATE_VARIABLE_LIST = ["rate_id", "specie_id", "process_id",
                      "ref",
                      "lhsA_id", "lhsB_id", "rhsA_id", "rhsB_id",
                      "threshold",
                      "wavelength",
                      "lhs_v", "rhs_v", "lhs_j", "rhs_j", "background",
                      "form"]
RATE_DTYPE = {"rate_id": int,
              "specie": str,
              "process": str,
              "ref": str,
              "lhs_a": str,
              "lhs_b": str,
              "rhs_a": str,
              "rhs_b": str,
              "threshold": float,
              "wavelength": float,
              "lhs_v": int,
              "rhs_v": int,
              "lhs_j": int,
              "rhs_j": int,
              "background": str,
              "form": str
              }
RATEDATA_VARIABLE_LIST = ["id", "rate_id", "num", "constant", "lau", "uau"]
RATEDATA_DTYPE = {"ratedata_id": int,
                  "num": int,
                  "constant": float,
                  "lau": float,
                  "uau": float}
MOD_DTYPE = {"model_name": str}

UPDATE_COMMAND_RATE = ("INSERT INTO rate " +
                       "SET "
                       "  rate_id = %s, "
                       "  ref = %s, "
                       "  threshold = %s, "
                       "  wavelength = %s, "
                       "  lhs_v = %s, "
                       "  rhs_v = %s, "
                       "  lhs_j = %s, "
                       "  rhs_j = %s, "
                       "  background = %s, "
                       "  form = %s, "
                       "  specie_id = (SELECT id FROM " + database + ".species "
                       "    WHERE NAME=%s), "
                       "  process_id = (select id from " + database + ".processes "
                       "    WHERE NAME=%s), "
                       "  lhsA_id = (select id from " + database + ".states "
                       "    WHERE NAME=%s), "
                       "  lhsB_id = (select id from " + database + ".states "
                       "    WHERE NAME=%s), "
                       "  rhsA_id = (select id from " + database + ".states "
                       "    WHERE NAME=%s), "
                       "  rhsB_id = (select id from " + database + ".states "
                       "    WHERE NAME=%s);")

INSERT_COMMAND_RATEDATA = insert_command("ratedata", RATEDATA_VARIABLE_LIST)

INSERT_COMMAND_MODELS2RATE = ("INSERT INTO models2rate "
                              "SET "
                              " rate_id=%s, "
                              " model_id=(SELECT model_id "
                              "           FROM " + database + ".models "
                              "           WHERE NAME=%s);")

file_number = 1
for directoryname in DIR_NAMES:
    directory = os.fsencode(NEPC_DATA + directoryname)
    for file in os.listdir(directory):
        if file_number >= MAX_RATE:
            print("WARNING: only processed " + str(file_number) +
                  " rate files. There appears to be addtional "
                  "data not processed.")
            break
        filename = os.fsdecode(file)
        if filename.endswith(".met") or filename.endswith(".mod"):
            continue
        else:
            # print("file_number: " + str(file_number))
            file_number = file_number + 1
            filename_wo_ext = filename.rsplit(".", 1)[0]
            mod_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".mod"])
            met_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".met"])

            dat_file = "".join([os.fsdecode(directory),
                                filename_wo_ext,
                                ".dat"])

            met_data = pd.read_csv(met_file,
                                   sep='\t',
                                   dtype=RATE_DTYPE,
                                   na_values=NA_VALUES)
            dat_data = pd.read_csv(dat_file,
                                   sep='\t',
                                   dtype=RATEDATA_DTYPE,
                                   na_values=NA_VALUES)

            F_RATE_DAT_FILE.write(
                "\t".join([str(met_data.iloc[0]['rate_id']),
                           directoryname + str(filename_wo_ext)]) + "\n"
            )

            MYCURSOR.execute(UPDATE_COMMAND_RATE,
                             (int(met_data.iloc[0]['rate_id']),
                              met_data.iloc[0]['ref'],
                              met_data.iloc[0]['threshold'],
                              met_data.iloc[0]['wavelength'],
                              int(met_data.iloc[0]['lhs_v']),
                              int(met_data.iloc[0]['rhs_v']),
                              int(met_data.iloc[0]['lhs_j']),
                              int(met_data.iloc[0]['rhs_j']),
                              met_data.iloc[0]['background'],
                              np_str(met_data, 0, 'form'),
                              np_str(met_data, 0, 'specie'),
                              np_str(met_data, 0, 'process'),
                              np_str(met_data, 0, 'lhs_a'),
                              np_str(met_data, 0, 'lhs_b'),
                              np_str(met_data, 0, 'rhs_a'),
                              np_str(met_data, 0, 'rhs_b')))
            dat_data.insert(1, 'rate_id', met_data.iloc[0]['rate_id'])
            dat_data_list = list(dat_data.itertuples(index=False,
                                                     name=None))
            MYCURSOR.executemany(INSERT_COMMAND_RATEDATA, dat_data_list)

            if os.path.exists(mod_file):
                mod_data = pd.read_csv(mod_file,
                                       sep='\t',
                                       dtype=MOD_DTYPE,
                                       na_values=NA_VALUES)
                mod_data.insert(0, 'rate_id', met_data.iloc[0]['rate_id'])
                mod_data_list = list(mod_data.itertuples(index=False,
                                                         name=None))
                MYCURSOR.executemany(INSERT_COMMAND_MODELS2RATE,
                                     mod_data_list)

            MYDB.commit()

MYDB.commit()
F_RATE_DAT_FILE.close()

if ARGS.debug:
    print_timestep("loaded data into rate, ratedata, and models2rate tables")
    # print_table("rate")

print_timestep("built " + database + " database")
for table in ["species", "processes", "states", "cs", "models",
              "models2cs", "csdata", "rate", "models2rate", "ratedata"]:
    print(table + ": " + str(nepc.count_table_rows(MYCURSOR, table)) +
          " rows")

####################
# Close connection #
####################

MYCURSOR.close()

MYDB.close()



from nepc import nepc
import pytest


@pytest.fixture
def nepc_connect(local, dbug):
    """Establishes a connection with the nepc_test database

    Parameters
    ----------
    local : boolean
        Checks whether the database is locally based or based off
        of 'ppdadamsonlinux'
    dbug : boolean
        Checks whether debug mode is on or off

    Returns
    -------
    cnx : MySQLConnection
        A connection to the official NEPC MySQL database
    cursor : MySQLCursor
        A MySQLCursor object for executing SQL queries
    """
    if dbug:
        print("opening database connection")
    cnx, cursor = nepc.connect(local, dbug, test=True)
    yield [cnx, cursor]
    if dbug:
        print("closing database connection")
    cursor.close()
    cnx.close()



"""Tests for nepc/nepc.py"""
import pandas as pd
import pytest
import mysql.connector
import nepc


@pytest.mark.usefixtures("nepc_connect")
def test_connect(nepc_connect):
    """Verify that nepc.connect() method connects to the NEPC database
    """
    assert isinstance(nepc_connect[0],
                      mysql.connector.connection_cext.CMySQLConnection)
    assert isinstance(nepc_connect[1],
                      mysql.connector.cursor_cext.CMySQLCursor)


@pytest.mark.usefixtures("nepc_connect")
def test_count_table_rows(nepc_connect):
    """Verify that nepc.count_table_rows returns an integer value"""
    rows = nepc.count_table_rows(nepc_connect[1], "species")
    assert isinstance(rows, int)


@pytest.mark.usefixtures("nepc_connect")
def test_cs_e_sigma(nepc_connect):
    """Verify that nepc.cs_e_sigma returns
    proper formats for the e_energy and sigma values"""
    # TODO: check some actual e, sigma values from the files
    e_energy, sigma = nepc.cs_e_sigma(nepc_connect[1], 1)
    assert isinstance(e_energy, list)
    assert isinstance(sigma, list)
    assert isinstance(e_energy[0], float)
    assert isinstance(sigma[0], float)


@pytest.mark.usefixtures("nepc_connect")
def test_cs_e(nepc_connect):
    """Verify that nepc.cs_e returns
    proper formats for the e_energy values for a single cs_id"""
    # TODO: check some actual e_energy values from the files
    e_energy = nepc.cs_e(nepc_connect[1], 1)
    assert isinstance(e_energy, list)
    assert isinstance(e_energy[0], float)


@pytest.mark.usefixtures("nepc_connect")
def test_cs_sigma(nepc_connect):
    """Verify that nepc.cs_sigma returns
    proper formats for the sigma values for a single cs_id"""
    # TODO: check some actual sigma values from the files
    sigma = nepc.cs_sigma(nepc_connect[1], 1)
    assert isinstance(sigma, list)
    assert isinstance(sigma[0], float)


@pytest.mark.usefixtures("nepc_connect")
def test_cs_metadata(nepc_connect):
    """Verify that nepc.cs_metadata returns the
    metadata as a list containing an int
    in the first index, a str in the second index,
    and a float in the fourth index"""
    metadata = nepc.cs_metadata(nepc_connect[1], 1)
    assert isinstance(metadata, list)
    assert isinstance(metadata[0], int)
    assert isinstance(metadata[1], str)
    assert isinstance(metadata[3], float)


@pytest.mark.usefixtures("nepc_connect")
def test_CS_class(nepc_connect):
    """Verify that the nepc.CS class has metadata and data attributes,
    and that each attribute is of the correct type"""
    # FIXME: randomly sample a few cross sections in the database
    cs = nepc.CS(nepc_connect[1], 1)
    # FIXME: add assert for each key in metadata
    assert isinstance(cs.metadata, dict)
    assert isinstance(cs.data, dict)
    assert isinstance(cs.metadata["cs_id"], int)
    assert isinstance(cs.metadata["specie"], str)
    assert isinstance(cs.metadata["units_e"], float)
    assert isinstance(cs.data["e"], list)
    assert isinstance(cs.data["e"][0], float)
    assert isinstance(cs.data["sigma"], list)
    assert isinstance(cs.data["sigma"][0], float)
    # test len()
    # FIXME: use length of data file to get length of sampled cross section
    assert len(cs) == 21


@pytest.mark.usefixtures("nepc_connect")
def test_CustomCS_class(nepc_connect):
    """Verify that the nepc.CustomCS class has metadata and data attributes,
    and that each attribute is of the correct type"""
    cs = nepc.CustomCS(nepc_connect[1], cs_id=1)
    # FIXME: add assert for each in key in metadata
    # FIXME: test edge cases of changing NEPC cross sections
    # FIXME: check that process, specie, state are in NEPC database
    assert isinstance(cs.metadata, dict)
    assert isinstance(cs.data, dict)
    assert cs.metadata["cs_id"] is None
    assert isinstance(cs.metadata["specie"], str)
    assert isinstance(cs.metadata["units_e"], float)
    assert isinstance(cs.data["e"], list)
    assert isinstance(cs.data["e"][0], float)
    assert isinstance(cs.data["sigma"], list)
    assert isinstance(cs.data["sigma"][0], float)
    cs = nepc.CustomCS(nepc_connect[1], cs_id=1,
                       metadata={'cs_id': -1,
                                 'specie': 'O2'},
                       data={'e': [0.0, 1.0, 2.0],
                             'sigma': [0.0, 0.1, 0.2]})
    assert isinstance(cs.metadata, dict)
    assert isinstance(cs.data, dict)
    assert cs.metadata["cs_id"] == -1
    assert cs.metadata["specie"] == 'O2'
    assert isinstance(cs.metadata["units_e"], float)
    assert isinstance(cs.data["e"], list)
    assert isinstance(cs.data["e"][0], float)
    assert isinstance(cs.data["sigma"], list)
    assert isinstance(cs.data["sigma"][0], float)
    assert cs.data['e'] == [0.0, 1.0, 2.0]
    assert cs.data['sigma'] == [0.0, 0.1, 0.2]
    """option not implemented
    cs = nepc.CustomCS(metadata={'cs_id': -1,
                                 'specie': 'N2',
                                 'process': 'excitation',
                                 'lhsA': 'N2(X1Sigmag+)',
                                 'rhsA': 'N2(W3Deltau)',
                                 'units_e': 1.0,
                                 'units_sigma': 1.0,
                                 'threshold': 7.36,
                                 'background': 'N2 W3DELTA-CARTWRIGHT 1977.'},
                       data={'e': [0.0, 1.0, 2.0],
                             'sigma': [0.0, 0.1, 0.2]})
    assert isinstance(cs.metadata, dict)
    assert isinstance(cs.data, dict)
    assert cs.metadata["cs_id"] == -1
    assert cs.metadata["specie"] == 'N2'
    assert cs.metadata["process"] == 'excitation'
    assert cs.metadata["lhsA"] == 'N2(X1Sigmag+)'
    assert cs.metadata["lhsB"] == '\\N'
    assert cs.metadata["rhsA"] == 'N2(X1Sigmag+)'
    assert cs.metadata["rhsB"] == '\\N'
    assert cs.metadata["background"] == 'N2 W3DELTA-CARTWRIGHT 1977.'
    assert cs.metadata["units_e"] == 1.0
    assert cs.metadata["units_sigma"] == 1.0
    assert cs.metadata["threshold"] == 7.36
    assert cs.metadata["lpu"] == -1
    assert cs.metadata["upu"] == -1
    assert isinstance(cs.metadata["units_e"], float)
    assert isinstance(cs.data["e"], list)
    assert isinstance(cs.data["e"][0], float)
    assert isinstance(cs.data["sigma"], list)
    assert isinstance(cs.data["sigma"][0], float)
    assert cs.data['e'] == [0.0, 1.0, 2.0]
    assert cs.data['sigma'] == [0.0, 0.1, 0.2]
    """


@pytest.mark.usefixtures("nepc_connect")
def test_Model_class(nepc_connect):
    """Verify that the nepc.Model class is a list of CS type and
    test the methods within the Class."""
    fict = nepc.Model(nepc_connect[1], "fict")
    # FIXME: randomly sample the models in the database
    # FIXME: randomly sample the cross sections in the model
    assert isinstance(fict.cs, list)
    assert isinstance(fict.cs[0].metadata, dict)
    assert isinstance(fict.cs[0].data, dict)
    assert isinstance(fict.cs[0].metadata['specie'], str)
    # test len method
    # FIXME use data in .mod files to determine number of CS in Model for assert
    assert len(fict) == 30

@pytest.mark.usefixtures("nepc_connect")
def test_CustomModel_class(nepc_connect):
    """Verify that the nepc.CustomModel class is a list of CS 
    and CustomCS type and
    test the methods within the Class."""
    fict = nepc.CustomModel(cursor=nepc_connect[1], model_name="fict")
    # FIXME: randomly sample the models in the database
    # FIXME: randomly sample the cross sections in the model
    assert isinstance(fict.cs, list)
    assert isinstance(fict.cs[0].metadata, dict)
    assert isinstance(fict.cs[0].data, dict)
    assert isinstance(fict.cs[0].metadata['specie'], str)
    assert len(fict.cs) == 30
    fict_mod = nepc.CustomModel(cursor=nepc_connect[1], 
            model_name="fict", 
            cs_id_list=[1, 2])
    # FIXME: randomly sample the models in the database
    # FIXME: randomly sample the cross sections in the model
    assert isinstance(fict_mod.cs, list)
    assert isinstance(fict_mod.cs[0].metadata, dict)
    assert isinstance(fict_mod.cs[0].data, dict)
    assert isinstance(fict_mod.cs[0].metadata['specie'], str)
    assert len(fict_mod.cs) == 32
    fict_subset = nepc.CustomModel(cursor=nepc_connect[1], 
            model_name="fict",
            metadata={'process': 'excitation'})
    # FIXME: randomly sample the models in the database
    # FIXME: randomly sample the cross sections in the model
    assert isinstance(fict_subset.cs, list)
    assert isinstance(fict_subset.cs[0].metadata, dict)
    assert isinstance(fict_subset.cs[0].data, dict)
    assert isinstance(fict_subset.cs[0].metadata['specie'], str)
    assert len(fict_subset.cs) == 15


@pytest.mark.usefixtures("nepc_connect")
def test_table_as_df(nepc_connect):
    """Verify when nepc.table_as_df is called it
    returns a DataFrame copy of what is included in the
    table used as an argument"""
    statef = nepc.table_as_df(nepc_connect[1], "states")
    assert isinstance(statef, pd.DataFrame)
    processf = nepc.table_as_df(nepc_connect[1], "processes",
                                columns=["id", "name"])
    assert isinstance(processf, pd.DataFrame)


@pytest.mark.usefixtures("nepc_connect")
def test_reaction_latex(nepc_connect):
    """Verify when nepc.reaction_latex is called it
    returns a string representing the LaTeX
    for the reaction from a nepc cross section"""
    # FIXME: verify latex is correct
    # FIXME: randomly sample cross sections
    for i in range(1, 30):
        cs = nepc.CS(nepc_connect[1], i)
        assert isinstance(nepc.reaction_latex(cs), str)


@pytest.mark.usefixtures("nepc_connect")
def test_model_summary_df(nepc_connect):
    """Verify nepc.model_summary_df returns a dataframe"""
    # TODO: check formatting of summary dataframe
    # FIXME: randomly sample models
    fict = nepc.Model(nepc_connect[1], "fict")
    df = fict.summary(lower=0, upper=10)
    assert isinstance(df, pd.io.formats.style.Styler)


@pytest.mark.usefixtures("nepc_connect")
def test_model_subset(nepc_connect):
    """Verify that Model.subset returns a proper
    subset of cross sections from a model within the NEPC MySQL database"""
    # TODO: test for lhsB and rhsB
    # FIXME: randomly sample models and cross sections within the model
    fict = nepc.Model(nepc_connect[1], "fict")
    fict_subset = fict.subset({'process': 'excitation',
                               'lhsA': 'N2(X1Sigmag+)'})
    assert isinstance(fict_subset, list)
    assert isinstance(fict_subset[0].metadata, dict)
    assert isinstance(fict_subset[0].data, dict)
    with pytest.raises(Exception):
        assert fict.subset()


@pytest.mark.usefixtures("nepc_connect")
def test_cs_subset_exception(nepc_connect):
    """When prompted with an exception in test_cs_subset
    verify that nepc.cs_subset does indeed return a value"""
    with pytest.raises(Exception):
        assert nepc.cs_subset(nepc_connect[1])



"""Utility methods for the nepc module.

"""
import sys
import shutil
import os
from subprocess import check_output


def wc_fxn(file_to_count):
    """Return the number of lines in a file using the commandline utility ``wc``.

    """
    return int(check_output(["wc", "-l", file_to_count]).split()[0])


def get_size(obj, seen=None):
    """Recursively find the size of an object.

    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def rmdir(outdir):
    """Try to remove a tree and raise an exception if this isn't possible.

    Parameters
    ----------
    outdir : str
        The directory to be removed.

    """
    try:
        shutil.rmtree(outdir)
    except OSError as exc:
        print("Error: %s - %s." % (exc.filename, exc.strerror))


def mkdir(outdir):
    """Try to make a directory and raise an exception if the directory already exists

    Parameters
    ----------
    outdir : str
        The directory to be made.

    """
    try:
        os.mkdir(outdir)
    except OSError as excep:
        print("Error: %s - %s." % (excep.filename, excep.strerror))



from nepc import nepc
from nepc.util import config
from nepc.util import util
import pandas as pd
import os
import pytest
import platform
# TODO: remove dependence on csv; put function in scraper that uses built-in
#       readlines function
import csv

# TODO: test that all values in [nepc]/tests/data are in the nepc database

NEPC_HOME = config.nepc_home()
NEPC_DATA = NEPC_HOME + "/tests/data/"
DIR_NAMES = [NEPC_HOME + "/tests/data/cs/n2/fict/",
             NEPC_HOME + "/tests/data/cs/n2/fict_total/"]


@pytest.mark.usefixtures("nepc_connect")
def test_csdata_lines(nepc_connect):
    cs_lines = 0
    for directoryname in DIR_NAMES:
        directory = os.fsencode(directoryname)

        for file in os.listdir(directory):
            filename = os.fsdecode(file)
            if filename.endswith(".met") or filename.endswith(".mod"):
                continue
            else:
                # subtract 1 to account for header
                cs_lines += util.wc_fxn(directoryname + filename) - 1

    assert cs_lines == nepc.count_table_rows(nepc_connect[1], "csdata")


@pytest.mark.usefixtures("nepc_connect")
def test_data_entered(nepc_connect, local):
    if local is False or platform.node() == 'ppdadamsonlinux':
        cs_dat_files = pd.read_csv(NEPC_DATA + 'cs_datfile_prod.tsv',
                                   delimiter='\t')
    else:
        cs_dat_files = pd.read_csv(NEPC_DATA + 'cs_datfile_local.tsv',
                                   delimiter='\t')

    for index, row in cs_dat_files.iterrows():
        cs_id = row['cs_id']
        dat_file = row['filename']
        df = pd.read_csv(NEPC_DATA + dat_file + '.dat', delimiter='\t',
                         usecols=['e_energy', 'sigma'])
        e_energy, sigma = nepc.cs_e_sigma(nepc_connect[1], cs_id)
        # assert e_energy == pytest.approx(df['e_energy'].tolist())
        assert sigma == pytest.approx(df['sigma'].tolist())


@pytest.mark.usefixtures("nepc_connect")
def test_meta_entered(nepc_connect, local, dbug):
    if local is False or platform.node() == 'ppdadamsonlinux':
        cs_dat_files = pd.read_csv(NEPC_DATA + 'cs_datfile_prod.tsv',
                                   delimiter='\t')
    else:
        cs_dat_files = pd.read_csv(NEPC_DATA + 'cs_datfile_local.tsv',
                                   delimiter='\t')

    for index, row in cs_dat_files.iterrows():
        cs_id = row['cs_id']
        met_file = row['filename']
        if dbug:
            print(cs_id, met_file)
        e, sigma = nepc.cs_e_sigma(nepc_connect[1], cs_id)

        meta_cols = ['cs_id', 'specie', 'process', 'units_e',
                     'units_sigma', 'ref', 'lhsA',
                     'lhsB', 'rhsA', 'rhsB', 'threshold', 'wavelength',
                     'lhs_v', 'rhs_v', 'lhs_j', 'rhs_j',
                     'background', 'lpu', 'upu']

        with open(NEPC_DATA + met_file + ".met", 'r', newline='') as f:
                reader = csv.reader(f, delimiter='\t')
                next(reader)
                meta_disk = list(reader)[0]
        meta_disk = [meta_disk[i] for i in list(range(len(meta_cols)))]
        for i in [3, 4, 10, 11, 17, 18]:
            meta_disk[i] = (float(meta_disk[i]) if meta_disk[i] != '\\N'
                            else meta_disk[i])

        for i in [0, 12, 13, 14, 15]:
            meta_disk[i] = (int(meta_disk[i]) if meta_disk[i] != '\\N'
                            else meta_disk[i])

        meta_db = [nepc.cs_metadata(nepc_connect[1], cs_id)[i]
                   for i in list(range(0, len(meta_cols)))]
        if dbug:
            print('meta_db: {}\t from {}'.format(meta_db, met_file))
        for i in range(len(meta_cols)):
            if dbug:
                print('meta_db[{}]: {}\t from {}'.format(str(i), str(meta_db[i]), met_file))
            if (type(meta_db[i]) is float):
                assert (pytest.approx(meta_disk[i]) ==
                        pytest.approx(meta_db[i]))
            elif meta_db[i] is None:
                assert meta_disk[i] == '\\N'
            else:
                assert meta_disk[i] == meta_db[i]



"""Get environment variables used by nepc scripts.
"""
import os


def user_home():
    """Returns the user's home directory.

    """
    return os.environ.get('HOME')


def nepc_home():
    """Returns the path to the nepc directory.

    """
    return os.environ.get('NEPC_HOME')


def nepc_cs_home():
    """Returns the path to the nepc_cs directory.

    """
    return os.environ.get('NEPC_CS_HOME')



from .nepc import connect
from .nepc import count_table_rows
from .nepc import model_cs_id_list
from .nepc import cs_e_sigma
from .nepc import cs_e
from .nepc import cs_sigma
from .nepc import cs_metadata
from .nepc import table_as_df
from .nepc import reaction_latex
from .nepc import CS
from .nepc import CustomCS
from .nepc import Model
from .nepc import CustomModel


# if somebody does "from somepackage import *", this is what they will
# be able to access:
__all__ = [
        'connect',
        'model_cs_id_list',
        'count_table_rows',
        'cs_e_sigma',
        'cs_e',
        'cs_sigma',
        'cs_metadata',
        'table_as_df',
        'reaction_latex',
        'CS',
        'CustomCS',
        'Model',
        'CustomModel'
        ]



"""Adapted from BOLOS (https://github.com/aluque/bolos). Adaptations:
 - replaced 'next()' with '__next__()' for Python 3
 - broke out large comment to add separate dictionary entries for ``process``,
   ``param``, ``species``, etc.

Notes
-----
This module contains the code required to parse BOLSIG+-compatible files.
To make the code re-usabe in other projects it is independent from the rest of
the BOLOS code.

Most users would only use the method :func:`parse` in this module, which is
documented below.

"""
import re
import numpy as np
from nepc.util import config as nepc_config

RE_NL = re.compile('\n')
def parse(filename, has_arg=True, debug=False):
    """ Parses a BOLSIG+ cross-sections file.

    Parameters
    ----------
    filename: str
        Name of a Bolsig+-compatible cross-sections file.

    has_arg: bool, optional
        Whether threshold values are present in each block

    Returns
    -------
    processes : list of dict
        A list with all processes, in dictionary form, included in the file.

    Notes
    -----
    This function does not return :class:`process.Process` instances so that
    the parser is independent of the rest of the code and can be re-used in
    other projects.  If you want to convert a process in dictionary form `d` to
    a :class:`process.Process` instance, use

    >>> process = process.Process(**d)

    """
    processes = []
    with open(filename) as fp:
        for line in fp:
            try:
                key = line.strip()
                fread = KEYWORDS[key]

                # If the key is not found, we do not reach this line.
                if debug:
                    print('New process of type {}'.format(key))

                d = fread(fp, has_arg, debug=debug)
                d['kind'] = key
                d['filename'] = filename
                # refactored comment to header
                # old code called header comment which collided with key comment
                d['comment'] = []

                # new code to add separate entries for process, param, etc
                header = [s.strip() for s in RE_NL.split(d['header'])]
                for i in range(len(header)):
                    if debug:
                        print('processing header[{:d}]: {}'.format(i, header[i]))
                    key, value = header[i].split(':',1)
                    key = key.lower().replace('.','')
                    value = value.lstrip()
                    if key == 'comment':
                        d['comment'].append(value)
                    else:
                        d[key] = value

                del d['header']
                d['comment'] = ' '.join(d['comment'])

                processes.append(d)

            except KeyError:
                pass

    if debug:
        print('Parsing complete. {:d} processes read.'.format(len(processes)))

    return processes


# BOLSIG+'s user guide saye that the separators must consist of at least five dashes
def _read_until_sep(fp, debug=False):
    """ Reads lines from fp until we find a separator line. 

    """
    RE_SEP = re.compile("-----+")
    lines = []
    for line in fp:
        if RE_SEP.match(line.strip()):
            break
        lines.append(line.strip())

    return lines


def _read_block(fp, has_arg=True, debug=False):
    """ Reads data of a process, contained in a block.
    ``has_arg`` indicates wether we have to read an argument line.

    """
    target = fp.__next__().strip()
    if has_arg:
        arg = fp.__next__().strip()
    else:
        arg = None

    header = "\n".join(_read_until_sep(fp, debug=debug))

    if debug:
        print('Read process {}'.format(target))
    data = np.loadtxt(_read_until_sep(fp, debug=debug)).tolist()

    return target, arg, header, data

#
# Specialized funcion for each keyword. They all return dictionaries with the
# relevant attibutes.
# 
def _read_momentum(fp, has_arg=True, debug=False):
    """ Reads a MOMENTUM or EFFECTIVE block.

    """
    target, arg, header, data = _read_block(fp, has_arg=has_arg, debug=debug)
    mass_ratio = float(arg.split()[0])
    d = dict(target=target,
             mass_ratio=mass_ratio,
             header=header,
             data=data)

    return d

RE_ARROW = re.compile('<?->')
def _read_excitation(fp, has_arg=True, debug=False):
    """ Reads an EXCITATION or IONIZATION block. 

    """
    target, arg, header, data = _read_block(fp, has_arg=has_arg, debug=debug)
    lhs, rhs = [s.strip() for s in RE_ARROW.split(target)]

    d = dict(target=lhs,
             product=rhs,
             header=header,
             data=data)

    if (has_arg):
        if '<->' in target.split():
            threshold, weight_ratio = float(arg.split()[0]), float(arg.split()[1])
            d['weight_ratio'] = weight_ratio
        else:
            threshold = float(arg.split()[0])

        d['threshold'] = threshold

    return d


def _read_attachment(fp, has_arg=False, debug=False):
    """ Reads an ATTACHMENT block.

    """
    target, arg, header, data = _read_block(fp, has_arg=False, debug=debug)

    d = dict(header=header,
             data=data,
             threshold=0.0)
    lr = [s.strip() for s in RE_ARROW.split(target)]

    if len(lr) == 2:
        d['target'] = lr[0]
        d['product'] = lr[1]
    else:
        d['target'] = target

    return d


KEYWORDS = {"MOMENTUM": _read_momentum, 
            "ELASTIC": _read_momentum, 
            "EFFECTIVE": _read_momentum,
            "EXCITATION": _read_excitation,
            "IONIZATION": _read_excitation,
            "ATTACHMENT": _read_attachment}


def write_data_to_file(data_array, filename, start_csdata_id):
    """Given an array of electron energies and corresponding cross sections,
    write the data to a file in the correct format for a NEPC database.

    Parameters
    ----------
    data_array : :class:`numpy.ndarray`
        A numpy array of values to be entered into the filenumpy array. Each row is of
        length 2 (e, sigma).
    filename: str
        Name of the file where values of ``data_array`` should be written.
    start_csdata_id : int
        The first ``csdata_id`` for the supplied data.

    Returns
    -------
    : int
        The next csdata_id to use.

    """
    csdata_id = start_csdata_id
    write_f = open(filename, "x")
    write_f.write("\t".join(['csdata_id', 'e_energy', 'sigma']) + "\n")
    for i in range(len(data_array)):
        write_f.write(str(csdata_id) + "\t" + str(data_array[i, 0])
                      + "\t" + str(data_array[i, 1]) + "\n")
        csdata_id = csdata_id + 1
    write_f.close()
    return csdata_id


def write_pd_data_to_file(data_frame, filename, start_csdata_id):
    """Given a dataframe of csdata, write this to a file in the correct format
    Parameters
    ----------
    data_frame: pandas DataFrame
    A DataFrame containing cross section data to be entered into the file

    filename: file
    Name of the file where values of data_frame should be entered

    start_cs_data_id : int
    The id where the data should be placed

    Return
    ------
    csdata_id
    The next csdata_id to use
    """
    csdata_id = start_csdata_id
    write_f = open(filename, "x")
    write_f.write("\t".join(['csdata_id', 'e_energy', 'sigma']) + "\n")
    for i in range(len(data_frame)):
        write_f.write(str(csdata_id) + "\t" + str(data_frame.iloc[i]['e_energy'])
                      + "\t" + str(data_frame.iloc[i]['sigma']) + "\n")
        csdata_id = csdata_id + 1
    write_f.close()
    return csdata_id


def write_metadata_to_file(filename, cs_id, specie, process,
                           units_e, units_sigma, ref='\\N',
                           lhs_a='\\N', lhs_b='\\N',
                           rhs_a='\\N', rhs_b='\\N', threshold='-1', wavelength='-1',
                           lhs_v=-1, rhs_v=-1, lhs_j=-1, rhs_j=-1,
                           background='\\N', lpu='-1', upu='-1'):
    """Given metadata for a cross section data set,
    write the metadata to a file in the correct format for a NEPC database.

    Parameters
    ----------
    filename: str
        Name of the file where values of ``data_array`` should be written.
    cs_id: int
        ``cs_id`` corresponding to the cross section in the database.
    specie: str
        The short name corresponding to the specie in the database.
    process: str
        The short name corresponding to the electron scattering process in the 
        database.
    units_e: float
        The units, in eV, of the electron energies.
    units_sigma: float
        The units, in :math:`m^2`, of the cross sections.
    ref: str
        The short name for the reference corresponding to the dataset.
    lhs_a: str
        The short name for the lhs_a state for the process.
    lhs_b: str
        The short name for the lhs_b state.
    rhs_a: str
        The short name for the rhs_a state.
    rhs_b: str
        The short name for the rhs_b state.
    threshold: float
        The threshold electron energy for the process.
    wavelength: float
        The wavelength of the photon associated with the process, if applicable.
    lhs_v: int
        The vibrational energy associated with the LHS state, if applicable.
    rhs_v: int
        The vibrational energy associated with the RHS state, if applicable.
    lhs_j: int
        The rotational energy associated with the LHS state, if applicable.
    rhs_j: int
        The rotational energy associated with the RHS state, if applicable.
    background: str
        Background information for the cross section data.
    lpu: float
        Lower percent uncertainty of the cross section data.
    upu: float
        Upper percent uncertainty of the cross section data.

    Returns
    -------
    : int
        The next cs_id to use.

    """
    write_met = open(filename, "x")
    write_met.write(
        "\t".join(
            ('cs_id',
             'specie',
             'process',
             'units_e',
             'units_sigma',
             'ref',
             'lhs_a',
             'lhs_b',
             'rhs_a',
             'rhs_b',
             'threshold',
             'wavelength',
             'lhs_v',
             'rhs_v',
             'lhs_j',
             'rhs_j',
             'background',
             'lpu',
             'upu')) + "\n")
    write_met.write(
        "\t".join(
            (str(cs_id),
             specie,
             process,
             str(units_e),
             str(units_sigma),
             ref,
             lhs_a,
             lhs_b,
             rhs_a,
             rhs_b,
             str(threshold),
             str(wavelength),
             str(lhs_v),
             str(rhs_v),
             str(lhs_j),
             str(rhs_j),
             background,
             str(lpu),
             str(upu))))
    write_met.close()
    return cs_id + 1


def write_cs_to_file(filename, data_array, cs_id, start_csdata_id, specie,
                     process, units_e, units_sigma, ref='\\N',
                     lhs_a='\\N', lhs_b='\\N', rhs_a='\\N', rhs_b='\\N',
                     threshold='-1',
                     wavelength='-1',
                     lhs_v=-1, rhs_v=-1, lhs_j=-1, rhs_j=-1,
                     background='\\N', lpu='-1', upu='-1'):
    """Write both the cross-section data and metadata to a file
    Parameters
    ----------
    Contains a combination of parameters from write_data_to_file and
    write_metadata_to_file"""
    next_csdata_id = write_data_to_file(data_array, filename+".dat",
                                        start_csdata_id)
    next_cs_id = write_metadata_to_file(filename+".met", cs_id, specie,
                                        process, units_e, units_sigma,
                                        ref, lhs_a, lhs_b, rhs_a, rhs_b,
                                        threshold, wavelength, lhs_v, rhs_v,
                                        lhs_j, rhs_j, background, lpu, upu)
    return (next_cs_id, next_csdata_id)




def write_next_id_to_file(
        next_cs_id, next_csdata_id, test=False):
    """Write out the next id's for the database to a file.

    Parameters
    ----------
    next_cs_id : int
        The next cs_id to use
    next_csdata_id: int
        The next csdata_id to use
    test: bool, optional
        If true, the command is run for the ``nepc_test`` database. Otherwise,
        it is run for the ``nepc`` database. (Default is the ``nepc`` database.)

    """
    if test:
        nepc_data_home = nepc_config.nepc_home() + '/tests/data/'
    else:
        nepc_data_home = nepc_config.nepc_cs_home() + '/data/'
    filename = nepc_data_home + "/next_id.tsv"
    id_file = open(filename, "w+")
    id_file.write(
        "\t".join(
            ('next_cs_id',
             'next_csdata_id')) + "\n")
    id_file.write(
        "\t".join(
            (str(next_cs_id),
             str(next_csdata_id))))
    id_file.close()


def write_models_to_file(filename, models_array):
    """Write model data to a file.

    Parameters
    ----------
    filename : str
        The name of the file where the model data should be stored.
    models_array: list of str
        A list of models that should be added to ``filename``.

    """
    model_f = open(filename, "x")
    model_f.write("model_name\n")
    for i in range(len(models_array)):
        model_f.write(models_array[i] + "\n")
    model_f.close()


def get_next_ids(test=False):
    """Get the next ids from a file

    Parameters
    ----------
    test: bool, optional
        If true, the command is run for the ``nepc_test`` database. Otherwise,
        it is run for the ``nepc`` database. (Default is the ``nepc`` database.)

    Returns
    -------
    : int
        Next cs_id to use
    : int
        Next csdata_id to use

    """
    if test:
        nepc_data_home = nepc_config.nepc_home() + '/tests/data/'
    else:
        nepc_data_home = nepc_config.nepc_cs_home() + '/data/'
    filename = nepc_data_home + "/next_id.tsv"
    with open(filename) as id_file:
        id_line = id_file.readlines()
    next_cs_id, next_csdata_id = id_line[1].split('\t')
    return int(next_cs_id), int(next_csdata_id)


def get_states(test=False):
    """Get lists of name's and long_name's from states.tsv file."""
    if test:
        nepc_data_home = nepc_config.nepc_home() + '/tests/data/'
    else:
        nepc_data_home = nepc_config.nepc_cs_home() + '/data/'
    filename = nepc_data_home + 'states.tsv'
    with open(filename) as states_f:
        states_lines = states_f.readlines()[1:]

    states = []
    for line in states_lines:
        states.append(line.split('\t'))
    return ([states[i][1] for i in range(len(states))],
            [states[i][2] for i in range(len(states))])


def remove_crs(mystring):
    """Removes new lines"""
    return mystring.replace('\n', ' ').replace('\r', '')


def text_array_to_float_array(text_array, omit_regexp=''):
    """Convert an array of strings to an array of floats

    Parameters
    ----------
    text_array : str array
    An array of strings to be converted

    Returns
    -------
    data : float array
    An array of floats containing float-converted data
    from text_array"""
    data = np.empty(shape=(0, 1))
    for inde in range(0, len(text_array)):
        if (omit_regexp == '' or not(
                bool(re.search(omit_regexp, text_array[inde])))):
            data = np.append(data, [[float(text_array[inde])]], axis=0)
    return data





"""This package provides the following functionality for NRL Evaluated Plasma
Chemistry (NEPC) style databases:

 - create data files for building a NEPC database from common sources (e.g. LXCat)
 - build a NEPC style database on a MySQL server
 - establishing a connection to a local or remote database
 - access cross section data via the CS class
 - access pre-defined plasma chemistry models via the Model class
 - curate, visualize, and use cross section data
 - perform exploratory data analysis (EDA) of cross section data
 - print statistics about a database (e.g. number of rows in various tables)

Examples
--------
Establish a connection to the database named `nepc` running on a
production server:

    >>> cnx, cursor = nepc.connect()

Establish a connection to the database named `nepc`
running on the local machine:

    >>> cnx, cursor = nepc.connect(local=True, test=True)

Access the pre-defined plasma chemistry model, `fict`, in the `nepc_test` database:

    >>> fict = nepc.Model(cursor, "fict")

Print a summary of the ``fict`` model, including a stylized Pandas dataframe:

    >>> fict.summary()

Additional examples of EDA using nepc are in ``tests/data/eda``. Examples of methods for
building data files for the ``nepc_test`` database, including parsing
`LXCat <https://nl.lxcat.net/data/set_type.php>`_ formatted data,
are in ``tests/data/methods``.

"""
import numpy as np
from pandas import DataFrame
import mysql.connector
import matplotlib.pyplot as plt


def connect(local=False, DBUG=False, test=False):
    """Establish a connection to a NEPC MySQL database

    Parameters
    ----------
    local : bool, optional
        Access a database on localhost; otherwise use the production
        server (default False).
    DBUG : bool, optional
        Print debug info (default False).
    test: bool, optional
        If true, access the `nepc_test` database; otherwise, connect to the `nepc` database.

    Returns
    -------
    cnx : `connection.MySQLConnection <https://dev.mysql.com/doc/connectors/en/connector-python-api-mysqlconnection.html>`_
        A connection to a NEPC MySQL database.
    cursor : `cursor.MySQLCursor <https://dev.mysql.com/doc/connectors/en/connector-python-api-mysqlcursor.html>`_
        A MySQLCursor object that can execute operations such as SQL statements. `cursor` 
        interacts with the NEPC server using the `cnx` connection.

    """
    if local:
        hostname = 'localhost'
    else:
        hostname = '132.250.158.124'

    if DBUG:  # pragma: no cover
        print("\nUsing NEPC database on " + hostname)

    if test:
        database = 'nepc_test'
    else:
        database = 'nepc'

    config = {'user': 'nepc',
              'password': 'nepc',
              'host': hostname,
              'database': database,
              'raise_on_warnings': True}

    cnx = mysql.connector.connect(**config)
    cursor = cnx.cursor()
    return cnx, cursor


def count_table_rows(cursor, table: str):
    """Return the number of rows in a MySQL table.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    table : str
        Name of a table in the NEPC database at ``cursor``.

    Returns
    -------
    : int
        Number of rows in ``table``.

    """
    cursor.execute("select count(*) from " + table + ";")
    table_rows = cursor.fetchall()
    return table_rows[0][0]


def model_cs_id_list(cursor, model_name):
    """Get a list of ``cs_id``'s for a model in a NEPC database.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    model_name : str
        Name of a model in the NEPC MySQL database

    Returns
    -------
    cs_id_list : list of int
        cs_id's corresponding to cross sections in the model

    """
    cursor.execute("SELECT cs.cs_id as cs_id " +
                   "FROM cs " +
                   "JOIN models2cs m2cs ON (cs.cs_id = m2cs.cs_id) " +
                   "JOIN models m ON (m2cs.model_id = m.model_id) " +
                   "WHERE m.name LIKE '" + model_name + "'")
    cs_id_list = cursor.fetchall()
    cs_id_list = [cs_id[0] for cs_id in cs_id_list]
    return cs_id_list


def cs_e_sigma(cursor, cs_id):
    """Get electron energy and cross section data for a given ``cs_id`` in a NEPC database.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        The ``cs_id`` for a cross section dataset in the NEPC database at ``cursor``.

    Returns
    -------
    : list of float
        Electron energies for the cross section dataset corresponding to ``cs_id``.
    : list of float
        Cross sections for the cross section dataset corresponding to ``cs_id``.

    """
    cursor.execute("SELECT e, sigma FROM csdata WHERE cs_id = " +
                   str(cs_id))
    cross_section = cursor.fetchall()
    # print(cross_section)
    e_energy = [i[0] for i in cross_section]
    sigma = [i[1] for i in cross_section]
    return e_energy, sigma


def cs_e(cursor, cs_id):
    """Get the electron energies for a cross section dataset in a NEPC database 
    corresponding to a given ``cs_id``.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        The ``cs_id`` for a cross section dataset in the NEPC database at ``cursor``.

    Returns
    -------
    : list of float
        Electron energies for the cross section dataset corresponding to ``cs_id``.

    """
    cursor.execute("SELECT e FROM csdata WHERE cs_id = " +
                   str(cs_id))
    cross_section = cursor.fetchall()
    # print(cross_section)
    return [i[0] for i in cross_section]


def cs_sigma(cursor, cs_id):
    """Get the cross sections for a cross section dataset in a NEPC database
    corresponding to a given ``cs_id``.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        The ``cs_id`` for a cross section dataset in the NEPC database at ``cursor``.

    Returns
    -------
    : list of float
        Cross sections for the cross section dataset corresponding to ``cs_id``.

    """
    cursor.execute("SELECT sigma FROM csdata WHERE cs_id = " +
                   str(cs_id))
    sigma = cursor.fetchall()
    # print(sigma)
    return [i[0] for i in sigma]


def cs_metadata(cursor, cs_id):
    """Get metadata for a given ``cs_id`` in a NEPC database.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        The ``cs_id`` for a cross section dataset in the NEPC database at ``cursor``.

    Returns
    -------
    list
        See :attr:`CS.metadata`. List items are in same order as :attr:`.CS.metadata`.

    """
    cursor.execute("SELECT A.`cs_id` , "
                   "B.`name` , "
                   "C.`name` , "
                   "A.`units_e`, A.`units_sigma`, A.`ref`, "
                   "D.`name`, E.`name`, "
                   "F.`name`, G.`name`, "
                   "A.`threshold`, A.`wavelength`, A.`lhs_v`, A.`rhs_v`, "
                   "A.`lhs_j`, A.`rhs_j`, "
                   "A.`background`, A.`lpu`, A.`upu`, "
                   "D.`long_name`, E.`long_name`, "
                   "F.`long_name`, G.`long_name`, "
                   "C.`lhs_e`, C.`rhs_e`, "
                   "C.`lhs_hv`, C.`rhs_hv`, "
                   "C.`lhs_v`, C.`rhs_v`, "
                   "C.`lhs_j`, C.`rhs_j` "
                   "FROM `cs` AS A "
                   "LEFT JOIN `species` AS B "
                   "ON B.`id` = A.`specie_id` "
                   "LEFT JOIN `processes` AS C "
                   "ON C.`id` = A.`process_id` "
                   "LEFT JOIN `states` AS D "
                   "ON D.`id` = A.`lhsA_id` "
                   "LEFT JOIN `states` AS E "
                   "ON E.`id` = A.`lhsB_id` "
                   "LEFT JOIN `states` AS F "
                   "ON F.`id` = A.`rhsA_id` "
                   "LEFT JOIN `states` AS G "
                   "ON G.`id` = A.`rhsB_id` "
                   "WHERE A.`cs_id` = " + str(cs_id))

    return list(cursor.fetchall()[0])


class CS:
    r"""A cross section data set, including metadata and cross section data,
    from a NEPC MySQL database.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        i.d. of the cross section in `cs` and `csdata` tables

    Attributes
    ----------
    metadata : dict
        cs_id : int
            id of the cross section in `cs` and `csdata` tables
        specie : str
            `name` of specie from `species` table
        process : str
            `name` of process from `processes` table
        units_e : float
            units of electron energy list e in eV
        units_sigma : float
            units of cross section list sigma in m^2
        ref : str
            `ref` from `cs` table corresponding to entry in
            '[nepc]/models/ref.bib'
        lhsA : str
            `name` of lhsA state from `states` table
        lhsB : str
            `name` of lhsB state from `states` table
        rhsA : str
            `name` of rhsA state from `states` table
        rhsB : str
            `name` of rhsB state from `states` table
        wavelength : float
            wavelength of photon involved in process in nanometers (nm)
        lhs_v : int
            vibrational energy level of lhs specie
        rhs_v : int
            vibrational energy level of rhs specie
        lhs_j : int
            rotational energy level of lhs specie
        rhs_j : int
            rotational energy level of rhs specie
        background : str
            background text describing origin of data and other important info
        lpu : float
            lower percent uncertainty
        upu : float
            upper percent uncertainty
        lhsA_long : str
            `long_name` of lhsA state from `states` table
        lhsB_long : str
            `long_name` of lhsB state from `states` table
        rhsA_long : str
            `long_name` of rhsA state from `states` table
        rhsB_long : str
            `long_name` of rhsB state from `states` table
        e_on_lhs : int
            number of electrons on lhs
        e_on_rhs : int
            number of electrons on rhs
        hv_on_lhs : int
            photon on lhs? (0 or 1)
        hv_on_rhs : int
            photon on rhs? (0 or 1)
        v_on_lhs : int
            vibrational energy level on lhs? (0 or 1)
        v_on_rhs : int
            vibrational energy level on rhs? (0 or 1)
        j_on_lhs : int
            rotational energy level on lhs? (0 or 1)
        j_on_rhs : int
            rotational energy level on rhs? (0 or 1)
    data : dict
        e : list of float
            Electron energies in units of ``units_e`` eV (see :attr:`.CS.metadata`).
        sigma : list of float
            Cross sections in units of ``units_sigma`` :math:`m^2` (see :attr:`.CS.metadata`).

    """
    def __init__(self, cursor, cs_id):
        metadata = cs_metadata(cursor, cs_id)
        self.metadata = {"cs_id": metadata[0],
                         "specie": metadata[1],
                         "process": metadata[2],
                         "units_e": metadata[3],
                         "units_sigma": metadata[4],
                         "ref": metadata[5],
                         "lhsA": metadata[6],
                         "lhsB": metadata[7],
                         "rhsA": metadata[8],
                         "rhsB": metadata[9],
                         "threshold": metadata[10],
                         "wavelength": metadata[11],
                         "lhs_v": metadata[12],
                         "rhs_v": metadata[13],
                         "lhs_j": metadata[14],
                         "rhs_j": metadata[15],
                         "background": metadata[16],
                         "lpu": metadata[17],
                         "upu": metadata[18],
                         "lhsA_long": metadata[19],
                         "lhsB_long": metadata[20],
                         "rhsA_long": metadata[21],
                         "rhsB_long": metadata[22],
                         "e_on_lhs": metadata[23],
                         "e_on_rhs": metadata[24],
                         "hv_on_lhs": metadata[25],
                         "hv_on_rhs": metadata[26],
                         "v_on_lhs": metadata[27],
                         "v_on_rhs": metadata[28],
                         "j_on_lhs": metadata[29],
                         "j_on_rhs": metadata[30]}

        e_energy, sigma = cs_e_sigma(cursor, cs_id)
        self.data = {"e": e_energy,
                     "sigma": sigma}


    def __len__(self):
        r"""The number of data points in the cross section data set"""
        return len(self.data['e'])


    def plot(self, units_sigma=1E-20, plot_param_dict={'linewidth': 1},
             xlim_param_dict={'auto': True}, ylim_param_dict={'auto': True},
             ylog=False, xlog=False, show_legend=True, filename=None,
             width=10, height=10):
        r"""Plot a single cross section data set.

        Parameters
        ----------
        units_sigma : float, optional
            desired units of the y-axis in :math:`m^2`
        plot_param_dict : dict, optional
            kwargs to pass to :meth:`matplotlib.axes.Axes.plot`
        xlim_param_dict: dict, optional
            kwargs to pass to :meth:`matplotlib.axes.Axes.set_xlim`
        ylim_param_dict: dict, optional
            kwargs to pass to :meth:`matplotlib.axes.Axes.set_ylim`
        ylog : bool, optional
            whether y-axis is log scale (default is False)
        xlog : bool, optional
            whether x-axis is log scale (default is False)
        show_legend: bool, optional
            whether to display the legend or not (default is True)
        filename: str, optional
            filename for output, if provided (default is to not output a file)
        width: float, optional
            width of plot
        height: float, optional
            height of plot

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Plot of the cross section data, :attr:`.CS.data`, with formatting
            using information in the metadata, :attr:`.CS.metadata`.

        """
        _, axes = plt.subplots()

        if ylog:
            plt.yscale('log')

        if xlog:
            plt.xscale('log')

        plt.rcParams["figure.figsize"] = (width, height)
        units_sigma_tex = "{0:.0e}".format(units_sigma) + " m$^2$"
        plt.ylabel(r'Cross Section (' + units_sigma_tex + ')')
        plt.xlabel(r'Electron Energy (eV)')

        axes.set_xlim(**xlim_param_dict)
        axes.set_ylim(**ylim_param_dict)

        axes.tick_params(direction='in', which='both',
                         bottom=True, top=True, left=True, right=True)

        reaction = reaction_latex(self)
        label_items = [self.metadata['process'], ": ", reaction]
        label_text = " ".join(item for item in label_items if item)
        e_np = np.array(self.data['e'])
        sigma_np = np.array(self.data['sigma'])

        upu = self.metadata['upu']
        lpu = self.metadata['lpu']
        if upu != -1:
            sigma_upper_np = sigma_np*(1 + upu)
            if lpu == -1:
                sigma_lower_np = sigma_np
        if lpu != -1:
            sigma_lower_np = sigma_np*(1 - lpu)
            if upu == -1:
                sigma_upper_np = sigma_np

        plot = axes.plot(e_np,
                         sigma_np*self.metadata['units_sigma']/units_sigma,
                         **plot_param_dict,
                         label='{}'.format(label_text))

        if upu != -1 or lpu != -1:
            fill_color = plot[0].get_color()
            axes.fill_between(e_np, sigma_lower_np, sigma_upper_np,
                              color=fill_color, alpha=0.4)

        if show_legend:
            axes.legend(fontsize=12, ncol=2, frameon=False,
                        bbox_to_anchor=(1.0, 1.0))
            # ax.legend(box='best',
            #           bbox_to_anchor=(0.5, 0.75), ncol=1, loc='center left')

        if filename is not None:
            plt.savefig(filename)

        return axes


class CustomCS(CS):
    """
    Extends :class:`.CS` to provide a custom cross section data set.

    If building upon an existing :class:`.CS`, must provide cursor and cs_id as well
    as one or both of metadata and data.

    If building a :class:`.CustomCS` from scratch, must provide :obj:`.metadata` 
    and :obj:`.data`.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    cs_id : int
        i.d. of the cross section in `cs` and `csdata` tables
    metadata : dict
        one or more of the attributes of :class:`.CS`
    data : dict
        same as attributes of :class:`.CS`

    Attributes
    ----------
    metadata : dict
        see Attributes of :class:`.CS`

    data : dict
        e : list[float]
            electron energy
        sigma : list[float]
            cross section

    """
    def __init__(self, cursor=None, cs_id=None, metadata=None, data=None):
        if ((cursor is None and cs_id is not None) or (cursor is not None and cs_id is None)):
            raise ValueError('If providing cursor or cs_id, must provide both.')
        if (cursor is not None and cs_id is not None):
            super().__init__(cursor, cs_id)
            self.metadata['cs_id'] = None
            if metadata is not None:
                for key in metadata.keys():
                    self.metadata[key] = metadata[key]
            if data is not None:
                self.data = data.copy()
        elif (data is None or metadata is None):
            raise ValueError('must provide data/metadata if not providing cursor/cs_id')
        else:
            self.metadata = metadata.copy()
            self.data = data.copy()


class Model:
    """A pre-defined collection of cross sections from a NEPC database

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    model_name :str
        Name of a NEPC model (pre-defined collection of cross sections)

    Attributes
    ----------
    cs: list[:class:`.CS`]
        cross section data in the NEPC format (:class:`.CS`)
    unique: list[float]
        set with :attr:`.Model.set_unique`, all unique electron energies in all :attr:`.CS.data`
        of the :class:`.Model`

    """
    def __init__(self, cursor, model_name):
        _cs_list = []
        _cs_id_list = model_cs_id_list(cursor, model_name)
        for cs_id in _cs_id_list:
            _cs_list.append(CS(cursor, cs_id))

        self.cs = _cs_list

    def __len__(self):
        """number of cross sections in the model"""
        return len(self.cs)

    def subset(self, metadata=None):
        """Select the cross sections in the model matching the provided metadata

        Parameters
        ----------
        metadata: dict
            see :attr:`.CS.metadata`

        Returns
        -------
        cs_subset: list[:class:`.CS`]
            cross section data in the NEPC format (:class:`.CS`)

        """
        if metadata is None or not isinstance(metadata, dict):
            raise Exception("must provide metadata of type dict")
        cs_subset = []
        for cs in self.cs:
            passed_filter = True
            for key in metadata.keys():
                if cs.metadata[key] != metadata[key]:
                    passed_filter = False
            if passed_filter:
                cs_subset.append(cs)
        return cs_subset


    def summary(self, metadata=None, lower=None, upper=None, sort=[]):
        """Summarize the NEPC model.

        Prints the following information:
            - Number of cross sections in the model
            - Number of cross sections matching metadata, if provided

        Returns a stylized Pandas dataframe with headers given by:

        headers = ["cs_id", "specie", "lhsA", "rhsA", "process",
                   "reaction", "threshold", "E_peak", "E_upper",
                   "sigma_max", "lpu", "upu"]

        Parameters
        ----------
        metadata: dict
            see :attr:`.CS.metadata`
        lower : int
            lower bound of model index to include in summary
        upper : int
            upper bound of model index to include in summary
        sort : list[str]
            headers by which the stylized Pandas table is sorted

        Returns
        -------
        cs_df : pandas.io.formats.style.Styler
            A stylized Pandas DataFrame containing the cs_id, process,
            range of electron energies (E_lower, E_upper),
            maximum sigma (sigma_max), and
            lpu/upu's for each cross section in the model (or subset of the
            model if :obj:`metadata` is provided)

        """
        summary_list = []

        headers = ["cs_id", "specie", "lhsA", "rhsA", "process",
                   "reaction", "threshold", "E_peak", "E_upper",
                   "sigma_max", "lpu", "upu"]

        max_e_peak = 0
        min_e_peak = 100000
        max_e_upper = 0
        max_peak_sigma = 0
        min_peak_sigma = 1
        max_lpu = 0.000000001
        max_upu = 0.000000001

        print('Number of cross sections in model: {:d}'.format(len(self.cs)))
        if metadata is not None:
            cs_subset = self.subset(metadata=metadata)
            print('Number of cross sections with '
                  'matching metadata: {:d}'.format(len(cs_subset)))
        else:
            cs_subset = self.cs


        for cs in cs_subset:
            csdata = np.array(list(zip(cs.data['e'], cs.data['sigma'])))
            e_peak = csdata[np.argmax(csdata[:, 1]), 0]
            cs_peak_sigma = np.max(csdata[:, 1])
            e_upper = np.max(csdata[csdata[:, 1] != 0.0][:, 0])
            if e_peak > max_e_peak:
                max_e_peak = e_peak
            if e_peak < min_e_peak:
                min_e_peak = e_peak
            if e_upper > max_e_upper:
                max_e_upper = e_upper
            if cs_peak_sigma > max_peak_sigma:
                max_peak_sigma = cs_peak_sigma
            if cs_peak_sigma < min_peak_sigma:
                min_peak_sigma = cs_peak_sigma
            reaction = reaction_latex(cs)
            cs_lpu = cs.metadata["lpu"]
            cs_upu = cs.metadata["upu"]
            if cs_lpu is not None and cs_lpu > max_lpu:
                max_lpu = cs_lpu
            if cs_upu is not None and cs_upu > max_upu:
                max_upu = cs_upu
            summary_list.append([cs.metadata["cs_id"],
                                 cs.metadata["specie"], cs.metadata["lhsA"], cs.metadata["rhsA"],
                                 cs.metadata["process"], reaction,
                                 cs.metadata["units_e"]*cs.metadata["threshold"],
                                 cs.metadata["units_e"]*e_peak,
                                 cs.metadata["units_e"]*e_upper,
                                 cs.metadata["units_sigma"]*cs_peak_sigma,
                                 cs_lpu, cs_upu])

        cs_df = DataFrame(summary_list, columns=headers)
        if sort:
            cs_df = (cs_df.sort_values(by=sort)
                     .reset_index(drop=True))
        if upper is None:
            upper = len(cs_df)
        if lower is None:
            lower = 0
        return (cs_df.loc[lower:upper]
                .style
                .background_gradient(subset=['threshold', 'E_peak', 'E_upper',
                                             'sigma_max', 'lpu', 'upu'],
                                     cmap='plasma')
                .highlight_null('red'))


    def set_unique(self):
        """sets :attr:`.Model.unique`

        """
        for cs, i in zip(self.cs, range(len(self.cs))):
            if i == 0:
                _unique = np.asarray(cs.data['e'])
            else:
                _unique = np.unique(np.concatenate([_unique, np.asarray(cs.data['e'])]))
        self.unique = list(_unique)


    def plot(self,
             units_sigma=1E-20,
             process='',
             plot_param_dict={'linewidth': 1},
             xlim_param_dict={'auto': True},
             ylim_param_dict={'auto': True},
             ylog=False, xlog=False, show_legend=True,
             filename=None,
             max_plots=10, width=10, height=10):
        """Plot cross section data in the Model.

        Parameters
        ----------
        process: str
            If provided, the process that should be plotted.
        units_sigma : float
            Desired units of the y-axis in m^2.
        plot_param_dict : dict
            kwargs to pass to ax.plot
        xlim(ylim)_param_dict: dict
            dictionary of kwargs to pass to ax.set_x(y)lim
        ylog, xlog: bool
            whether y-, x-axis is log scale
        show_legend: bool
            whether to display the legend or not
        filename: str
            filename for output, if provided (default is to not output a file)
        max_cs : int
            maximum number of CS to put on graph

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Plot of the cross section data, :attr:`.CS.data`, with formatting
            using information in the metadata, :attr:`.CS.metadata` within the
            model.

        """
        _, axes = plt.subplots()

        if ylog:
            plt.yscale('log')

        if xlog:
            plt.xscale('log')

        plt.rcParams["figure.figsize"] = (width, height)
        units_sigma_tex = "{0:.0e}".format(units_sigma) + " m$^2$"
        plt.ylabel(r'Cross Section (' + units_sigma_tex + ')')
        plt.xlabel(r'Electron Energy (eV)')

        axes.set_xlim(**xlim_param_dict)
        axes.set_ylim(**ylim_param_dict)

        axes.tick_params(direction='in', which='both',
                         bottom=True, top=True, left=True, right=True)

        plot_num = 0
        for i in range(len(self.cs)):
            if plot_num >= max_plots:
                continue
            elif process in ('', self.cs[i].metadata['process']):
                plot_num += 1

                reaction = reaction_latex(self.cs[i])
                label_items = [self.cs[i].metadata['process'], ": ", reaction]
                label_text = " ".join(item for item in label_items if item)
                e_np = np.array(self.cs[i].data['e'])
                sigma_np = np.array(self.cs[i].data['sigma'])

                upu = self.cs[i].metadata['upu']
                lpu = self.cs[i].metadata['lpu']
                if upu != -1:
                    sigma_upper_np = sigma_np*(1 + upu)
                    if lpu == -1:
                        sigma_lower_np = sigma_np
                if lpu != -1:
                    sigma_lower_np = sigma_np*(1 - lpu)
                    if upu == -1:
                        sigma_upper_np = sigma_np

                plot = axes.plot(e_np,
                                 sigma_np*self.cs[i].metadata['units_sigma']/units_sigma,
                                 **plot_param_dict,
                                 label='{}'.format(label_text))

                if upu != -1 or lpu != -1:
                    fill_color = plot[0].get_color()
                    axes.fill_between(e_np, sigma_lower_np, sigma_upper_np,
                                      color=fill_color, alpha=0.4)

        if show_legend:
            axes.legend(fontsize=12, ncol=2, frameon=False,
                        bbox_to_anchor=(1.0, 1.0))
            # ax.legend(box='best',
            #           bbox_to_anchor=(0.5, 0.75), ncol=1, loc='center left')

        if filename is not None:
            plt.savefig(filename)

        return axes


class CustomModel(Model):
    """Extends :class:`.Model` to provide a customized collection of cross sections.

    The cross sections can be of class :class:`.CS`
    (from the NEPC database) or :class:`.CustomCS` (user created/modified).
    Options:

    a. If building upon an existing NEPC model, must provide cursor and model_name. May
       also provide a filter to select cross sections that meet criteria.
    b. If building from existing cross sections, must provide cursor and cs_id_list.
    c. If building from a list of custom cross sections, must provide cs_list.

    Must do at least one of a, b, or c, and can do any combination thereof.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    model_name :str
        The name of a NEPC model (see [nepc.wiki]/models
    cs_id_list : list of int
        List of cs_id's to pull from NEPC database.
    cs_list : list of CustomCS
        List of user-defined cross sections of CustomCS type.
    metadata: dict
        Dictionary of filter criteria to select specific CS's from a Model.

    Attributes
    ----------
    cs: list of :class:`.CS` or :class:`.CustomCS`
        A list of cross section data and metadata of CS or CustomCS type.

    """
    def __init__(self, cursor=None, model_name=None, cs_id_list=[], cs_list=[], metadata=None):
        if model_name is None and not cs_id_list and not cs_list:
            raise ValueError('Must provide at least one of model_name, cs_id_list, or cs_list')

        if cursor is None and (model_name is not None or cs_id_list):
            raise ValueError('Must provide cursor if providing model_name or cs_id_list')

        if cursor is not None and (model_name is None and not cs_id_list):
            raise ValueError('Must provide model_name or cs_id_list if providing cursor.')

        if metadata is not None and (cursor is None or model_name is None):
            raise ValueError('Must provide model_name and cursor if providing metadata.')

        if cs_list:
            _cs_list = cs_list.copy()
        else:
            _cs_list = []

        if cs_id_list:
            _cs_id_list = cs_id_list.copy()
            for cs_id in _cs_id_list:
                _cs_list.append(CS(cursor, cs_id))

        if model_name is not None:
            _model_cs_id_list = model_cs_id_list(cursor, model_name)
            for cs_id in _model_cs_id_list:
                cs = CS(cursor, cs_id)
                if metadata is not None:
                    passed_filter = True
                    for key in metadata.keys():
                        if cs.metadata[key] != metadata[key]:
                            passed_filter = False
                if metadata is None or passed_filter:
                    _cs_list.append(cs)

        self.cs = _cs_list.copy()


def table_as_df(cursor, table, columns="*"):
    """Return a ``table`` in a MySQL database as a pandas DataFrame.

    Parameters
    ----------
    cursor : cursor.MySQLCursor
        A MySQLCursor object. See return value ``cursor`` of :func:`.connect`.
    table : str
        Name of a table in the NEPC database at ``cursor``.
    columns : list of str, optional
        Which columns to get. (Default is to get all columns.)

    Returns
    -------
    DataFrame
        Table in the form of a pandas DataFrame

    """
    column_text = ", ".join(columns)
    cursor.execute("SELECT " + column_text + " FROM " + table)
    return DataFrame(cursor.fetchall())


def reaction_latex(cs):
    """Return the LaTeX for the process involved in a nepc cross section.

    Parameters
    ----------
    cs : :class:`.CS` or :class:`.CustomCS`
        A nepc cross section.

    Returns
    -------
    : str
        The LaTeX for the process involved in a NEPC cross section.

    """
    # FIXME: move this method to the CS Class
    # FIXME: allow for varying electrons and including hv, v, j on rhs and lhs
    # FIXME: decide how to represent total cross sections and implement
    e_on_lhs = cs.metadata['e_on_lhs']
    if e_on_lhs == 0:
        lhs_e_text = None
    elif e_on_lhs == 1:
        lhs_e_text = "e$^-$"
    else:
        lhs_e_text = str(e_on_lhs) + "e$^-$"

    e_on_rhs = cs.metadata['e_on_rhs']
    if e_on_rhs == 0:
        rhs_e_text = None
    elif e_on_rhs == 1:
        rhs_e_text = "e$^-$"
    else:
        rhs_e_text = str(e_on_rhs) + "e$^-$"

    lhsA_text = cs.metadata['lhsA_long']
    if cs.metadata['process'] == 'excitation_v':
        lhsA_text = lhsA_text.replace(")", " v=" + str(cs.metadata['lhs_v']) + ")")
    lhsB_text = cs.metadata['lhsB_long']
    lhs_items = [lhs_e_text,
                 lhsA_text,
                 lhsB_text]
    lhs_text = " + ".join(item for item in lhs_items if item)

    rhsA_text = cs.metadata['rhsA_long']
    if cs.metadata['process'] == 'excitation_v':
        rhsA_text = rhsA_text.replace(")", " v=" + str(cs.metadata['rhs_v']) + ")")
    rhsB_text = cs.metadata['rhsB_long']
    rhs_items = [rhsA_text,
                 rhsB_text,
                 rhs_e_text]
    rhs_text = " + ".join(item for item in rhs_items if item)
    reaction = " $\\rightarrow$ ".join([lhs_text, rhs_text])
    return reaction



"""Utilities for scraping data from PDFs and other docs"""
import re
import numpy as np
from IPython.display import display
import pdfplumber


def get_pdf(pdf_filename):
    """Returns an easily extractable version of a pdf file

    Parameters
    ----------
    pdf_filename : str
        The pdf file which will be opened

    Returns
    -------
    pdfplumber.open(pdf_filename)
        The extractable version of pdf_filename

    """
    return pdfplumber.open(pdf_filename)


def get_column_data(pdf, page_number, crop_dim_array,
                    locate_tables=False, omit_regex=''):
    pg = pdf.pages[page_number]
    data = np.empty(shape=(0, 1))
    for i in range(len(crop_dim_array)):
        pgCropped = pg.crop(crop_dim_array[i])
        if (locate_tables):
            display(pgCropped.to_image().reset().draw_rects(pgCropped.chars))
        else:
            textSpaceDelim = pgCropped.extract_text().split("\n")
            for j in range(len(textSpaceDelim)):
                if (omit_regex == '' or not(
                        bool(re.search(omit_regex, textSpaceDelim[j])))):
                    data = np.append(
                        data, [[float(textSpaceDelim[j])]],
                        axis=0)
    return data


def get_column_strings(pdf, page_number, crop_dim_array,
                       locate_tables=False, omit_regex=''):
    pg = pdf.pages[page_number]
    data = []
    pgCropped = pg.crop(crop_dim_array)
    if (locate_tables):
        display(pgCropped.to_image().reset().draw_rects(pgCropped.chars))
    else:
        textSpaceDelim = pgCropped.extract_text().split("\n")
        for j in range(len(textSpaceDelim)):
            if (omit_regex == '' or not(
                    bool(re.search(omit_regex, textSpaceDelim[j])))):
                # textArray = textSpaceDelim[j].split(' ')
                # data.append([textArray[0], textArray[1]])
                data.append(textSpaceDelim[j])
    return data


def get_table_data(pdf,
                   page_number, crop_dim_array,
                   locate_tables=False, omit_regex=''):
    """Get the table data from a pdf

    Parameters
    ----------
    pdf : pdf
        The pdf file from which data will be extracted

    page_number : int
        The page number from which the data should be extracted

    Returns
    -------
    data : N-dimensional array
        An N-dimensional array containing all of the table data
        from the pdf"""
    pages = pdf.pages[page_number]
    # data = [[] for i in range(len(crop_dim_array))]
    data = np.empty(shape=(0, 2))
    for i in range(len(crop_dim_array)):
        page_cropped = pages.crop(crop_dim_array[i])
        if (locate_tables):
            display(page_cropped.to_image().reset().draw_rects(
                page_cropped.chars))
        else:
            text_space_delim = page_cropped.extract_text().split("\n")
            for j in range(len(text_space_delim)):
                if (omit_regex == '' or not(
                        bool(re.search(omit_regex, text_space_delim[j])))):
                    text_array = text_space_delim[j].split(' ')
                    data = np.append(data, [[float(text_array[0]),
                                             float(text_array[1])]], axis=0)
    return data
